#include <stdio.h>

#include "md5.h"
#include "newinfi.h"

static const char magic[] = "niwa no mannaka no ki";

struct random_ctx {
  struct md5_ctx ctxi, ctx0, ctx;
  unsigned char digest[16];
  int b;
};

static void
random_init (struct random_ctx *rc,
	     const struct md5_ctx *ctxi)
{
  rc->ctxi = *ctxi;
  rc->ctx0 = *ctxi;
  rc->b = 0;
}

static unsigned char
random_byte (struct random_ctx *rc)
{
  size_t i;

  if ( rc->b == 16 )
    {
      MD5_feed (&rc->ctx0, '\n');
      rc->b = 0;
    }
  if ( rc->b == 0 )
    {
      rc->ctx = rc->ctx0;
      for ( i=0 ; magic[i] ; i++ )
	MD5_feed (&rc->ctx, magic[i]);
      MD5_feed (&rc->ctx, '\n');
      MD5_stop (&rc->ctx, rc->digest);
    }
  return rc->digest[rc->b++];
}

static unsigned long
random_long (struct random_ctx *rc)
{
  unsigned long w;

  w = random_byte (rc);
  w = (w<<8) | random_byte (rc);
  w = (w<<8) | random_byte (rc);
  w = (w<<8) | random_byte (rc);
  return w;
}

static void
random_init_from_cpath (struct random_ctx *rc,
			const char cpath[], const char extra[])
{
  struct md5_ctx ctxi;
  size_t i;

  MD5_start (&ctxi);
  for ( i=0 ; cpath[i] ; i++ )
    MD5_feed (&ctxi, cpath[i]);
  for ( i=0 ; extra[i] ; i++ )
    MD5_feed (&ctxi, extra[i]);
  MD5_feed (&ctxi, '\n');
  random_init (rc, &ctxi);
}

static size_t
canonical_path (char buf[], size_t buf_size, size_t *buf_len,
		const char path[], size_t path_len)
{
  static const char hexdigits[16] = "0123456789abcdef";
  size_t i;
  char b, c, cc;
  unsigned char ch;
  size_t len;

  len = 0;
  if ( len+1 < buf_size )
    buf[len++] = '/';
  b = 0;
  c = 1;
  for ( i=0 ; i<path_len ; i++ )
    {
      cc = 0;
      ch = 0;
      if ( path[i] == '/' )
	{
	  if ( b )
	    {
	      if ( len+1 < buf_size )
		buf[len++] = '/';
	    }
	  b = 0;
	  c = 1;
	}
      else if ( path[i] == ';' )
	{
	  c = 0;
	}
      else if ( path[i] == '%' )
	{
	  ch = 0;
	  if ( i+1 < path_len && path[i+1] >= '0' && path[i+1] <= '9' )
	    ch = path[i+1] - '0';
	  else if ( i+1 < path_len && path[i+1] >= 'A' && path[i+1] <= 'F' )
	    ch = path[i+1] - 'A' + 10;
	  else if ( i+1 < path_len && path[i+1] >= 'a' && path[i+1] <= 'f' )
	    ch = path[i+1] - 'a' + 10;
	  ch <<= 4;
	  if ( i+2 < path_len && path[i+2] >= '0' && path[i+2] <= '9' )
	    ch += path[i+2] - '0';
	  else if ( i+2 < path_len && path[i+2] >= 'A' && path[i+2] <= 'F' )
	    ch += path[i+2] - 'A' + 10;
	  else if ( i+2 < path_len && path[i+2] >= 'a' && path[i+2] <= 'f' )
	    ch += path[i+2] - 'a' + 10;
	  cc = c;
	  i += 2;
	}
      else
	{
	  ch = path[i];
	  cc = c;
	}
      if ( cc )
	{
	  if ( ( ch >= 'A' && ch <= 'Z' )
	       || ( ch >= 'a' && ch <= 'z' )
	       || ( ch >= '0' && ch <= '9' )
	       || ch == '-' || ch == '_' || ch == '.' || ch == '!'
	       || ch == '~' || ch == '*' || ch == '\''
	       || ch == '(' || ch == ')' )
	    {
	      if ( len+1 < buf_size )
		buf[len++] = ch;
	    }
	  else
	    {
	      if ( len+1 < buf_size )
		buf[len++] = '%';
	      if ( len+1 < buf_size )
		buf[len++] = hexdigits[ch>>4];
	      if ( len+1 < buf_size )
		buf[len++] = hexdigits[ch&0xf];
	    }
	  b = 1;
	}
    }
  if ( b )
    {
      if ( len+1 < buf_size )
	buf[len++] = '/';
    }
  buf[len] = 0;
  if ( buf_len )
    *buf_len = len;
  return len;
}

static const char *syllables[71] = {
  "a", "i", "u", "e", "o",
  "ka", "ki", "ku", "ke", "ko",
  "ga", "gi", "gu", "ge", "go",
  "sa", "shi", "su", "se", "so",
  "za", "zhi", "zu", "ze", "zo",
  "ta", "chi", "tsu", "te", "to",
  "da", "ji", "dzu", "de", "do",
  "na", "ni", "nu", "ne", "no",
  "ha", "hi", "fu", "he", "ho",
  "ba", "bi", "bu", "be", "bo",
  "pa", "pi", "pu", "pe", "po",
  "ma", "mi", "mu", "me", "mo",
  "ya", "yu", "yo",
  "ra", "ri", "ru", "re", "ro",
  "wa", "wo",
  "n"
};

static const char *links[10][13] = {
  {
    "aries/", "taurus/", "gemini/", "cancer/", "leo/", "virgo/",
    "libra/", "scorpius/", "sagittarius/", "capricornus/", "aquarius/", "pisces/",
    NULL
  },
  {
    "sol/", "luna/", "mercurius/", "venus/", "mars/", "iuppiter/",
    "saturnus/", "uranus/", "neptunus/", "pluto/",
    NULL
  },
  {
    "a/", "b/", "c/", NULL
  },
  {
    "*/", "**/", "***/", NULL
  },
  {
    "i/", "ii/", "iii/", "iv/", "v/", "vi/", "vii/", NULL
  },
  {
    "1/", "2/", "3/", "4/", "5/", "6/",
    "7/", "8/", "9/", "10/", "11/", "12/", NULL
  },
  {
    "ego/", "tu/", "id/", NULL
  },
  {
    "MAIN/", "SUB/", "SEC/", "IND/", NULL
  },
  {
    "do/", "re/", "mi/", "fa/", "so/", "la/", "ti/", NULL
  },
  {
    "Ahm/", "Mu/", "Ra/", "Beh/", "Summ/", "Cah/", "Lum/", "Om/", NULL
  },
};

static size_t
do_nonsense (char buf[], size_t buf_size, size_t *buf_len,
	     struct random_ctx *rc)
{
  size_t len;
  int nw, ns;
  int i, j;

  len = 0;
  nw = 1 + random_long (rc) %6;
  for ( i=0 ; i<nw ; i++ )
    {
      ns = 1 + random_long (rc) %4;
      for ( j=0 ; j<ns ; j++ )
	{
	  len += rsprintf (buf+len, buf_size-len, NULL,
			   ((j<ns-1||i==nw-1)?"%s":"%s "),
			   syllables[random_long (rc) %71]);
	}
    }
  if ( buf_len )
    *buf_len = len;
  return len;
}

void
do_body (char buf[], size_t buf_size, size_t *buf_len,
	 const char path[], size_t path_len)
{
  size_t len;
  struct random_ctx rc;
  char cpath[3*MAXREQSIZE];  /* Leave room for %-escaping! */
  size_t cpath_len;
  char nonsense[100];
  int np, ns;
  int k, l;
  char sp;
  size_t i, j;

  canonical_path (cpath, sizeof(cpath), &cpath_len, path, path_len);
  random_init_from_cpath (&rc, cpath, "");
  do_nonsense (nonsense, sizeof(nonsense), NULL, &rc);
  len = rsprintf (buf, buf_size, NULL,
		  "<?xml version=\"1.0\" encoding=\"us-ascii\"?>\r\n\r\n"
		  "<!DOCTYPE html PUBLIC "
		  "\"-//W3C//DTD XHTML 1.0 Strict//EN\" "
		  "\"http://www.w3.org/TR/xhtml1/DTD/"
		  "xhtml1-strict.dtd\">\r\n\r\n"
		  "<html xmlns=\"http://www.w3.org/1999/xhtml\">\r\n"
		  "<head>\r\n"
		  "<title>%s</title>\r\n"
		  "<meta name=\"robots\" content=\"noindex,nofollow\" />\r\n"
		  "</head>\r\n"
		  "<body>\r\n"
		  "<h1>%s</h1>\r\n", nonsense, nonsense);
  np = 1 + random_long (&rc) %6 + random_long (&rc) %6;
  for ( k=0 ; k<np ; k++ )
    {
      len += rsprintf (buf+len, buf_size-len, NULL,
		       "<p>");
      ns = 1 + random_long (&rc) %4 + random_long (&rc) %8;
      for ( l=0 ; l<ns ; l++ )
	{
	  do_nonsense (nonsense, sizeof(nonsense), NULL, &rc);
	  len += rsprintf (buf+len, buf_size-len, NULL,
			   "%s", nonsense);
	  if ( random_long (&rc) %3 == 0 )
	    {
	      len += rsprintf (buf+len, buf_size-len, NULL, ".");
	      sp = 2;
	    }
	  else if ( random_long (&rc) %3 == 0 && l < ns-1 )
	    {
	      len += rsprintf (buf+len, buf_size-len, NULL, ",");
	      sp = 1;
	    }
	  else if ( random_long (&rc) %3 == 0 && l < ns-1 )
	    {
	      len += rsprintf (buf+len, buf_size-len, NULL, ":");
	      sp = 1;
	    }
	  else if ( random_long (&rc) %3 == 0 && l < ns-1 )
	    {
	      len += rsprintf (buf+len, buf_size-len, NULL, ";");
	      sp = 1;
	    }
	  else if ( random_long (&rc) %3 == 0 )
	    {
	      len += rsprintf (buf+len, buf_size-len, NULL, "?");
	      sp = 2;
	    }
	  else if ( random_long (&rc) %3 == 0 )
	    {
	      len += rsprintf (buf+len, buf_size-len, NULL, "!");
	      sp = 2;
	    }
	  else
	    {
	      len += rsprintf (buf+len, buf_size-len, NULL, "&hellip;");
	      sp = 2;
	    }
	  if ( l < ns-1 )
	    {
	      if ( sp == 1 )
		len += rsprintf (buf+len, buf_size-len, NULL, " ");
	      else
		len += rsprintf (buf+len, buf_size-len, NULL, "  ");
	    }
	}
      len += rsprintf (buf+len, buf_size-len, NULL,
		       "</p>\r\n");
    }
  k = random_long (&rc) %10;
  len += rsprintf (buf+len, buf_size-len, NULL,
		   "<ul>\r\n");
  for ( l=0 ; links[k][l] ; l++ )
    {
      random_init_from_cpath (&rc, cpath, links[k][l]);
      do_nonsense (nonsense, sizeof(nonsense), NULL, &rc);
      len += rsprintf (buf+len, buf_size-len, NULL,
		       "<li><a href=\"%s%s\">%s</a></li>\r\n",
		       cpath, links[k][l], nonsense);
    }
  len += rsprintf (buf+len, buf_size-len, NULL,
		   "</ul>\r\n");
  j = 0;
  for ( i=0 ; cpath[i] ; i++ )
    {
      if ( cpath[i] == '/' && i<cpath_len-1 )
	{
	  j = i;
	}
    }
  cpath[j+1] = 0;
  cpath_len = j+1;
  random_init_from_cpath (&rc, cpath, "");
  do_nonsense (nonsense, sizeof(nonsense), NULL, &rc);
  len += rsprintf (buf+len, buf_size-len, NULL,
		   "<hr />\r\n<p>(<a href=\"%s\">%s</a>)</p>\r\n",
		   cpath, nonsense);
  len += rsprintf (buf+len, buf_size-len, NULL,
		   "</body>\r\n"
		   "</html>\r\n");
  if ( buf_len )
    *buf_len = len;
}
