#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>
#include <errno.h>

#include "newinfi.h"

struct slot scoreboard[MAXTHREADS];
pthread_mutex_t score_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t score_cond = PTHREAD_COND_INITIALIZER;
static unsigned long seqid0, seqid1;

static void *
thread_main (void *slot0)
{
  struct slot *slot = slot0;
  char inbuf[MAXREQSIZE];
  size_t inbuf_len;
  char outbuf[MAXOUTSIZE];
  size_t outbuf_len;
  char body[MAXOUTSIZE];
  size_t body_len;
  int v;
  struct req r;
  struct uri u;

  /* Read request. */
  v = read_request (slot->csock, inbuf, sizeof(inbuf), &inbuf_len);
  if ( v < 0 )
    make_internal_error (outbuf, sizeof(outbuf), &outbuf_len,
			 strerror (errno), 1);
  else if ( v == 1 )
    make_request_timeout (outbuf, sizeof(outbuf), &outbuf_len, 1);
  else if ( v == 2 )
    make_request_too_long (outbuf, sizeof(outbuf), &outbuf_len, 1);
  else if ( v == 3 )
    goto closeit;  /* Socket is closed. */
  else if ( v == 0 )
    {
      /* Have a request. */
      if ( parse_request (inbuf, inbuf_len, &r) < 0 )
	make_bad_request (outbuf, sizeof(outbuf), &outbuf_len, 1);
      else if ( parse_uri (inbuf+r.uri_start, r.uri_len, &u) < 0 )
	make_bad_uri (outbuf, sizeof(outbuf), &outbuf_len, r.method!=HEAD);
      else if ( r.method != GET && r.method != HEAD )
	make_unimplemented (outbuf, sizeof(outbuf), &outbuf_len, 1);
#if 1
      else if ( u.path_len == 11 && memcmp (inbuf+r.uri_start+u.path_start,
					    "/robots.txt", 11) == 0 )
	make_robots_txt (outbuf, sizeof(outbuf), &outbuf_len, r.method!=HEAD);
#endif
      else
	{
	  /* Generate page. */
	  do_body (body, sizeof(body), &body_len,
		   inbuf+r.uri_start+u.path_start, u.path_len);
	  make_200 (outbuf, sizeof(outbuf), &outbuf_len,
		    r.method!=HEAD, body, "html");
	}
    }
  /* Send response. */
  write_response (slot->csock, outbuf, outbuf_len, NULL);
 closeit:
  close (slot->csock);  /* Ignore errors! */
  /* Now signal that slot is free. */
  v = pthread_mutex_lock (&score_mutex);
  if ( v )
    {
      fprintf (stderr, "pthread_mutex_lock: %s\n",
	       strerror (v));
      exit (EXIT_FAILURE);
    }
  slot->inuse = 0;
  v = pthread_mutex_unlock (&score_mutex);
  if ( v )
    {
      fprintf (stderr, "pthread_mutex_unlock: %s\n",
	       strerror (v));
      exit (EXIT_FAILURE);
    }
  pthread_cond_signal (&score_cond);
  if ( v )
    {
      fprintf (stderr, "pthread_cond_signal: %s\n",
	       strerror (v));
      exit (EXIT_FAILURE);
    }
  return 0;
}

void
accept_loop (void)
{
  pthread_attr_t attr;
  int csock;
#ifdef USE_IPV6
  struct sockaddr_in6 rsa;
#else /* ndef USE_IPV6 */
  struct sockaddr_in rsa;
#endif /* ndef USE_IPV6 */
  socklen_t rsa_len = sizeof(rsa);
  int i;
  struct timeval start;
  struct timespec stop;
  int v;

  /* Create all threads detacheable. */
  v = pthread_attr_init (&attr);
  if ( v )
    {
      fprintf (stderr, "pthread_attr_init: %s\n",
	       strerror (v));
      exit (EXIT_FAILURE);
    }
  v = pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);
  if ( v )
    {
      fprintf (stderr, "pthread_attr_setdetachstate: %s\n",
	       strerror (v));
      exit (EXIT_FAILURE);
    }
  /* Main loop proper */
  while ( 1 )
    {
      /* Accept connection. */
      csock = accept (sock, (struct sockaddr *)&rsa, &rsa_len);
      if ( csock < 0 )
	{
	  perror ("accept");
	  if ( errno != EINTR )
	    exit (EXIT_FAILURE);
	  continue;
	}
      /* Determine sequence number. */
      seqid0 = (seqid0+1) & 0xFFFFFFFFUL;
      if ( ! seqid0 )
	seqid1 = seqid1+1;
      /* Now enter connection in scoreboard... */
      v = pthread_mutex_lock (&score_mutex);
      if ( v )
	{
	  fprintf (stderr, "pthread_mutex_lock: %s\n",
		   strerror (v));
	  exit (EXIT_FAILURE);
	}
      while ( 1 )
	{
	  for ( i=0 ; i<MAXTHREADS && scoreboard[i].inuse ; i++ ) ;
	  if ( i<MAXTHREADS )
	    /* Found available slot. */
	    {
	      pthread_t thr;

	      scoreboard[i].inuse = 1;
	      scoreboard[i].pos = i;
	      scoreboard[i].seqid0 = seqid0;
	      scoreboard[i].seqid1 = seqid1;
	      scoreboard[i].csock = csock;
	      scoreboard[i].rsa = rsa;
	      scoreboard[i].rsa_len = rsa_len;
	      v = pthread_mutex_unlock (&score_mutex);
	      if ( v )
		{
		  fprintf (stderr, "pthread_mutex_unlock: %s\n",
			   strerror (v));
		  exit (EXIT_FAILURE);
		}
	      /* Create worker thread. */
	      v = pthread_create (&thr, &attr, thread_main,
				  (void *)(scoreboard+i));
	      if ( v )
		{
		  fprintf (stderr, "pthread_create: %s\n",
			   strerror (v));
		  if ( v != EINTR && v != EAGAIN )
		    exit (EXIT_FAILURE);
		  /* We might be out of resources: try again later. */
		  v = pthread_mutex_lock (&score_mutex);
		  if ( v )
		    {
		      fprintf (stderr, "pthread_mutex_lock: %s\n",
			       strerror (v));
		      exit (EXIT_FAILURE);
		    }
		  scoreboard[i].inuse = 0;
		}
	      else
		break;  /* Return to accepting connections. */
	    }
	  /* At this stage, mutex is locked. */
	  /* Wait for a thread to finish (or 5s to elapse) and try
	   * again. */
	  if ( gettimeofday (&start, NULL) < 0 )
	    {
	      perror ("gettimeofday");
	      exit (EXIT_FAILURE);
	    }
	  stop.tv_sec = start.tv_sec + 5;
	  stop.tv_nsec = start.tv_usec*1000;
	  pthread_cond_timedwait (&score_cond, &score_mutex, &stop);
	  if ( v )
	    {
	      fprintf (stderr, "pthread_cond_timedwait: %s\n",
		       strerror (v));
	      if ( v != ETIMEDOUT && v != EINTR && v != EAGAIN )
		exit (EXIT_FAILURE);
	    }
	}
    }
}
