#ifndef _NEWINFI_H
#define _NEWINFI_H

#ifndef PORTNUM
#define PORTNUM 8091 /* TCP port to bind to */
#endif

#ifndef MAXTHREADS
#define MAXTHREADS 5 /* Maximal number of simultaneous threads */
#endif

#ifndef MAXOUTSIZE
#define MAXOUTSIZE 8192 /* Maximal size of reply */
#endif

#ifndef MAXREQSIZE
#define MAXREQSIZE 8192 /* Maximal size of request */
#endif

#ifndef MAXDELAY
#define MAXDELAY 60 /* Timeout delay */
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <pthread.h>

extern int sock;  /* The Master socket */
#ifdef USE_IPV6
extern struct sockaddr_in6 lsa; /* Local socket address */
#else /* ndef USE_IPV6 */
extern struct sockaddr_in lsa; /* Local socket address */
#endif /* ndef USE_IPV6 */
extern const socklen_t lsa_len; /* Size of above */

struct slot {
  char inuse; /* Is a thread alive in this slot? */
  int pos; /* scoreboard[pos].pos = pos */
  unsigned long seqid0, seqid1; /* Sequence number of request being served */
  int csock; /* Client socket */
#ifdef USE_IPV6
  struct sockaddr_in6 rsa; /* Remote socket address */
#else /* ndef USE_IPV6 */
  struct sockaddr_in rsa; /* Remote socket address */
#endif /* ndef USE_IPV6 */
  socklen_t rsa_len; /* Size of above */
};
extern struct slot scoreboard[MAXTHREADS];
extern pthread_mutex_t score_mutex;
extern pthread_cond_t score_cond;

struct req {
  enum http_method {
    OPTIONS, GET, HEAD, POST, PUT, DELETE, TRACE, CONNECT
  } method;
  size_t uri_start, uri_len, uri_stop;
};

struct uri {
  char scheme;
  size_t scheme_start, scheme_len, scheme_stop;
  size_t auth_start, auth_len, auth_stop;
  size_t path_start, path_len, path_stop;
  char query;
  size_t query_start, query_len, query_stop;
};

#if defined(__GNUC__)
size_t rsprintf (char buf[], size_t buf_size, size_t *buf_len,
		 const char template[], ...) __attribute__((__format__(__printf__,4,5)));
#else
size_t rsprintf (char buf[], size_t buf_size, size_t *buf_len,
		 const char template[], ...);
#endif
size_t rscpy (char buf[], size_t buf_size, size_t *buf_len,
	      const char str[], size_t str_len);
void do_bind (void);
void accept_loop (void);
int read_request (int csock, char buf[], size_t buf_size, size_t *buf_len);
int write_response (int csock, const char buf[], size_t buf_len,
		    size_t *buf_wlen);
void make_internal_error (char buf[], size_t buf_size, size_t *buf_len,
			  const char errstr[], char body);
void make_unimplemented (char buf[], size_t buf_size, size_t *buf_len,
			 char body);
void make_bad_request (char buf[], size_t buf_size, size_t *buf_len,
		       char body);
void make_bad_uri (char buf[], size_t buf_size, size_t *buf_len,
		   char body);
void make_request_timeout (char buf[], size_t buf_size, size_t *buf_len,
			   char body);
void make_request_too_long (char buf[], size_t buf_size, size_t *buf_len,
			    char body);
void make_200 (char buf[], size_t buf_size, size_t *buf_len,
	       char body, const char content[], const char typestr[]);
void make_robots_txt (char buf[], size_t buf_size, size_t *buf_len,
		      char body);
int parse_request (const char buf[], size_t buf_len, struct req *r);
int parse_uri (const char uri[], size_t uri_len, struct uri *u);
void do_body (char buf[], size_t buf_size, size_t *buf_len,
	      const char path[], size_t path_len);

#endif /* not defined _NEWINFI_H */
