#include <string.h>

#include "newinfi.h"

static int
match (const char matchstr[],
       const char buf[], size_t buf_len, size_t *mlen)
{
  size_t len;

  len = strlen (matchstr);
  if ( buf_len >= len && memcmp (buf, matchstr, len) == 0 )
    {
      if ( mlen )
	*mlen = len;
      return 1;
    }
  else
    return 0;
}

int
parse_request (const char buf[], size_t buf_len, struct req *r)
{
  char b, d;
  size_t uri_start, uri_stop;
  size_t i;

  if ( match ("OPTIONS ", buf, buf_len, &uri_start) )
    r->method = OPTIONS;
  else if ( match ("GET ", buf, buf_len, &uri_start) )
    r->method = GET;
  else if ( match ("HEAD ", buf, buf_len, &uri_start) )
    r->method = HEAD;
  else if ( match ("POST ", buf, buf_len, &uri_start) )
    r->method = POST;
  else if ( match ("PUT ", buf, buf_len, &uri_start) )
    r->method = PUT;
  else if ( match ("DELETE ", buf, buf_len, &uri_start) )
    r->method = DELETE;
  else if ( match ("TRACE ", buf, buf_len, &uri_start) )
    r->method = TRACE;
  else if ( match ("CONNECT ", buf, buf_len, &uri_start) )
    r->method = CONNECT;
  else return -1;
  r->uri_start = uri_start;
  b = 0;
  for ( i=uri_start ; i<buf_len ; i++ )
    {
      if ( buf[i] == ' ' )
	{
	  b = 1;
	  break;
	}
      else if ( buf[i] < 0x20 || buf[i] >= 0x7f )
	return -1;
    }
  if ( ! b )
    return -1;
  uri_stop = i;
  r->uri_stop = uri_stop;
  r->uri_len = uri_stop-uri_start;
  if ( ! match (" HTTP/", buf+uri_stop, buf_len-uri_stop, &i) )
    return -1;
  i += uri_stop;
  b = 0;
  d = 0;
  for ( ; i<buf_len ; i++ )
    {
      if ( buf[i] >= '0' && buf[i] <= '9' )
	d = 1;
      else if ( buf[i] == '.' )
	{
	  b = 1;
	  break;
	}
      else
	return -1;
    }
  if ( ! b || ! d )
    return -1;
  i++;
  b = 0;
  d = 0;
  for ( ; i<buf_len ; i++ )
    {
      if ( buf[i] >= '0' && buf[i] <= '9' )
	d = 1;
      else if ( buf[i] == '\r' || buf[i] == '\n' )
	{
	  b = 1;
	  break;
	}
      else
	return -1;
    }
  if ( ! b || ! d )
    return -1;
  b = 0;
  for ( ; i<buf_len ; i++ )
    {
      if ( buf[i] == '\r' )
	;
      else if ( buf[i] == '\n' )
	{
	  b = 1;
	  break;
	}
      else
	return -1;
    }
  if ( ! b )
    return -1;
  return 0;
}

int
parse_uri (const char uri[], size_t uri_len, struct uri *u)
{
  char b, a;
  size_t auth_start, path_start;
  size_t i;

  /* Check general character validity */
  for ( i=0 ; i<uri_len ; i++ )
    if ( uri[i] <= 0x20 || uri[i] >= 0x7f || uri[i] == '<' || uri[i] == '>'
	 || uri[i] == '#' || uri[i] == '"' )
      return -1;
    else if ( uri[i] == '%' )
      {
	if ( i+2 >= uri_len )
	  return -1;
	else if ( ! ( ( uri[i+1] >= 'A' && uri[i+1] <= 'F' )
		      || ( uri[i+1] >= 'a' && uri[i+1] <= 'f' )
		      || ( uri[i+1] >= '0' && uri[i+1] <= '9' ) ) )
	  return -1;
	else if ( ! ( ( uri[i+2] >= 'A' && uri[i+2] <= 'F' )
		      || ( uri[i+2] >= 'a' && uri[i+2] <= 'f' )
		      || ( uri[i+2] >= '0' && uri[i+2] <= '9' ) ) )
	  return -1;
	i += 2;
      }
  /* Find scheme */
  if ( uri_len > 0 && uri[0] == '/' )
    {
      u->scheme = 0;
      i = 0;
    }
  else
    {
      b = 0;
      a = 0;
      for ( i=0 ; i<uri_len ; i++ )
	{
	  if ( ( uri[i] >= 'A' && uri[i] <= 'Z' )
	       || ( uri[i] >= 'a' && uri[i] <= 'z' ) )
	    {
	      if ( i == 0 )
		a = 1;
	    }
	  else if ( ( uri[i] >= '0' && uri[i] <= '9' )
		    || uri[i] == '+' || uri[i] == '-' || uri[i] == '.' )
	    ;
	  else if ( uri[i] == ':' )
	    {
	      b = 1;
	      break;
	    }
	  else
	    return -1;
	}
      if ( ! b )
	return -1;
      u->scheme = 1;
      u->scheme_start = 0;
      u->scheme_stop = i;
      u->scheme_len = i;
      if ( i != 4 || ( uri[0] != 'h' && uri[0] != 'H' )
	   || ( uri[1] != 't' && uri[1] != 'T' )
	   || ( uri[2] != 't' && uri[2] != 'T' )
	   || ( uri[3] != 'p' && uri[3] != 'P' ) )
	return -1;
      /* Find authority */
      if ( i+2 >= uri_len || uri[i+1] != '/' || uri[i+2] != '/' )
	return -1;
      i += 3;
      auth_start = i;
      for ( ; i<uri_len ; i++ )
	{
	  if ( uri[i] == '/' || uri[i] == '?' )
	    break;
	}
      u->auth_start = auth_start;
      u->auth_stop = i;
      u->auth_len = i-auth_start;
    }
  /* Find path and query */
  path_start = i;
  for ( ; i<uri_len ; i++ )
    {
      if ( uri[i] == '?' )
	break;
    }
  u->path_start = path_start;
  u->path_stop = i;
  u->path_len = i-path_start;
  if ( i < uri_len && uri[i] == '?' )
    {
      i++;
      u->query = 1;
      u->query_start = i;
      u->query_stop = uri_len;
      u->query_len = uri_len-i;
    }
  else u->query = 0;
  return 0;
}
